from flask import Blueprint, json


def build_status_blueprint(ll):
    from liveline import LiveLine

    liveline: LiveLine = ll
    status = Blueprint('status', __name__)

    # pylint: disable=W0612
    @status.route('/health')
    def health_check():
        return json.jsonify({'health': 'ok'})

    # pylint: disable=W0612
    @status.route('/')
    def status_check():
        return json.jsonify({
            'running': not liveline.paused
        })

    return status


import json


from typing import List
from .order import Order
from .position import Position

class Portfolio:
    def __init__(self):
        pass

    def to_dict(self):
        return {
            'orders': [o.to_dict() for o in self.orders],
            'positions': [p.to_dict() for p in self.positions]
        }


    @property
    def orders(self) -> List[Order]:
        pass


    @property
    def positions(self) -> List[Position]:
        pass



from enum import Enum
from typing import Generic, TypeVar

from .equity import Equity

E = TypeVar('E', bound=Equity)


class OrderType(Enum):
    MARKET = 0
    LIMIT = 1
    STOP = 2

    def __str__(self):
        return self.name


class OrderStatus(Enum):
    PENDING = 0
    FULFILLED = 1
    CANCELED = 2
    FAILED = 3

    def __str__(self):
        return self.name


# pylint: disable=E1136
class Order(Generic[E]):
    def __init__(self,
                 equity: E,
                 quantity,
                 price,
                 order_type: OrderType=OrderType.MARKET,
                 order_status: OrderStatus=OrderStatus.PENDING):
        self._equity = equity
        self._quantity = quantity
        self._price = price
        self._order_type = order_type
        self._order_status = order_status

    def to_dict(self):
        return {
            'equity': self.equity.to_dict(),
            'price': '%.2f' % self.price,
            'quantity': self.quantity,
            'type': str(self.order_type),
            'status': str(self.order_status)
        }

    def __repr__(self):
        return '[$%.2f %s %s x%d]' % (self.price, self.equity, self.order_type, self.quantity)

    @property
    def equity(self) -> E:
        return self._equity

    @property
    def quantity(self) -> int:
        return self._quantity

    @property
    def price(self) -> float:
        return self._price

    @property
    def order_type(self) -> OrderType:
        return self._order_type

    @property
    def order_status(self) -> OrderStatus:
        return self._order_status



from .account import Account
from .equity import Equity, Stock, ETF, Options, OptionsUnderlyingEquity, OptionsType
from .quote import Quote
from .order import Order, OrderStatus, OrderType
from .portfolio import Portfolio
from .position import Position



from .equity import Equity


class Quote:
    def __init__(self, equity: Equity, ask_price, ask_size, bid_price, bid_size):
        self._equity = equity
        self._ask_price = ask_price
        self._ask_size = ask_size
        self._bid_price = bid_price
        self._bid_size = bid_size

    def __repr__(self):
        return '"%s @ $%.2f/$%.2f"' % (self.equity, self.ask_price, self.bid_price)

    @property
    def equity(self) -> Equity:
        return self._equity

    @property
    def ask_price(self) -> float:
        return self._ask_price

    @property
    def bid_price(self) -> float:
        return self._bid_price

    @property
    def ask_size(self) -> int:
        return self._ask_size

    @property
    def bid_size(self) -> int:
        return self._bid_size



import copy
import json
import logging
import os
import time
import traceback
from datetime import datetime, tzinfo
from threading import Thread, Timer
from typing import Dict, List

import arrow
import pandas as pd
import pandas_market_calendars as mcal
import pytz
import schedule
from flask import Flask

from liveline.data import LiveLineData
from liveline.model import Options, Order, OrderStatus, Stock
from liveline.order import LiveLineOrder, OrderExecutor, PoolExecutor
from liveline.web import (build_executions_blueprint, build_overview_blueprint,
                          build_status_blueprint, build_strategy_blueprint)

from .context import Context, ContextDataStore
from .execution_data_store import ExecutionDataStore


class Strategy:
    def __init__(self):
        self.logger = logging.getLogger(self.name)

    @property
    def name(self):
        pass

    @property
    def interval(self) -> int:
        """
        Interval to run this strategy in minutes.
        """
        return 5

    def run(self,
            context: Context,
            data: LiveLineData,
            order: LiveLineOrder):
        pass


class LiveLine:
    def __init__(self, data: LiveLineData, order: LiveLineOrder, context_store: ContextDataStore, execution_store: ExecutionDataStore):
        self.logger: logging.Logger = logging.getLogger('LiveLine')
        self.data: LiveLineData = data
        self.order: LiveLineOrder = order
        self.context_store = context_store
        self.execution_store = execution_store
        self.paused = True
        self.strategies: Dict[str, Strategy] = dict()

    @property
    def livelineorder(self) -> LiveLineOrder:
        return self.order

    @property
    def livelinedata(self) -> LiveLineData:
        return self.data

    @property
    def market_open_time(self) -> arrow.Arrow:
        return arrow.get(datetime(2018, 6, 1, 9, 30), 'US/Eastern')

    @property
    def market_close_time(self) -> arrow.Arrow:
        return arrow.get(datetime(2018, 6, 1, 16, 0), 'US/Eastern')

    def market_time_now(self) -> arrow.Arrow:
        return arrow.now('US/Eastern')

    def is_market_open_today(self):
        us_east_date = self.market_time_now()
        us_east_date = us_east_date.replace(hour=10, minute=1)

        nyse = mcal.get_calendar('NYSE')
        early = nyse.schedule(start_date='2018-07-01', end_date='2022-07-10')
        return nyse.open_at_time(early, pd.Timestamp(us_east_date.format('YYYY-MM-DD HH:mm'), tz='America/New_York'))

    def add_strategy(self, strategy: Strategy):
        if strategy.name in self.strategies:
            raise ValueError(
                'Strategy called %s already exist, cannot have duplicate names' % strategy.name)
        self.strategies[strategy.name] = strategy

    def get_strategies(self) -> List[Strategy]:
        return copy.copy(self.strategies)

    def __register_flask_blueprint__(self):
        status_bp = build_status_blueprint(self)
        self.app.register_blueprint(status_bp, url_prefix='/status')

        strategy_bp = build_strategy_blueprint(self)
        self.app.register_blueprint(strategy_bp, url_prefix='/strategy')

        executions_bp = build_executions_blueprint(self)
        self.app.register_blueprint(executions_bp, url_prefix='/execution')

        overview_bp = build_overview_blueprint(self)
        self.app.register_blueprint(overview_bp, url_prefix='/')

    def start(self, debug=False, web_server=True):
        if not debug:
            ll_thread = Thread(target=self.__schedule__,
                               name='Livelive engine')
            self.logger.info('===============================')
            self.logger.info('| LiveLine Engine has started |')
            self.logger.info('|        by Daniel Wang       |')
            self.logger.info('===============================')
            ll_thread.start()
        else:
            debug_thread = Thread(target=self.__day_start__, kwargs={"debug": debug})
            self.logger.debug('Running a debug liveline thread...')
            debug_thread.start()

        if web_server:
            self.app = Flask(__name__)
            flask_logging_level = logging.WARNING
            # pylint: disable=E1101
            self.app.logger.setLevel(flask_logging_level)
            logging.getLogger('werkzeug').setLevel(flask_logging_level)

            # register blueprints
            self.__register_flask_blueprint__()

            # start flask thread
            port = int(os.environ['PORT']) if 'PORT' in os.environ else 8080
            self.logger.info('Flask server listening at %d ...' % port)
            self.app.run(host='0.0.0.0', port=port)

    def __schedule__(self):
        local_market_open_time = self.market_open_time.to('local')
        local_market_close_time = self.market_close_time.to('local')
        engine_start_time = '%02d:%02d' % (local_market_open_time.hour,
                                           local_market_open_time.minute)
        engine_pause_time = '%02d:%02d' % (local_market_close_time.hour,
                                           local_market_close_time.minute)
        schedule.every().day.at(engine_start_time).do(self.__day_start__)
        schedule.every().day.at(engine_pause_time).do(self.__day_pause__)

        self.logger.info('Engine runs at %s everyday' % engine_start_time)
        self.logger.info('Engine pauses at %s everyday' % engine_pause_time)

        # trigger first run if within trade hour
        now = self.market_time_now().time()
        if now > self.market_open_time.time() and now < self.market_close_time.time():
            self.logger.info('Within trading hour, running engine now')
            self.__day_start__()

        while True:
            schedule.run_pending()
            time.sleep(1)

    def __day_start__(self, debug=False):
        if not debug and not self.is_market_open_today():
            self.logger.info("Market closed today, won't run")
            return

        self.logger.info('Engine started on market open or restarted')
        self.paused = False

        for name in self.strategies:
            try:
                self.__run_strategy__(self.strategies[name])
            except Exception as e:
                self.logger.error('Error when running strategy %s' %
                                  name, exc_info=e, stack_info=True)

    def __day_pause__(self):
        self.logger.info('Engine paused on market close')
        self.paused = True

    def __run_strategy__(self, strategy: Strategy):
        if self.paused:
            return

        # schedule a timer for next run
        interval_in_seconds = 60 * strategy.interval
        timer = Timer(interval_in_seconds, self.__run_strategy__, [strategy])
        timer.start()

        # run current loop for stategy
        order_placed = []
        self.logger.info('>> Running strategy %s' % strategy.name)
        try:
            # find context from previous execution
            context = self.context_store.get_strategy_latest_context(
                strategy.name)
            if context is None:
                self.logger.debug(
                    'No previous context found for %s, create a new one' % strategy.name)
                context = Context()
            else:
                self.logger.debug(
                    'Found a previous context for %s' % strategy.name)

            # build order pool for this execution
            stock_pool: PoolExecutor = PoolExecutor[Stock]()
            options_pool: PoolExecutor = PoolExecutor[Options]()
            ll_order_pool = LiveLineOrder(stock_pool, options_pool)

            # run strategy
            strategy.run(context, self.livelinedata, ll_order_pool)

            # save context
            try:
                self.context_store.save(strategy.name, context)
                self.logger.debug('Context saved for %s' % strategy.name)
            except Exception as e:
                self.logger.error('Cannot save context for %s' % strategy.name)
                raise e

            # validate and execute orders
            for stock_order in stock_pool.pool:
                self.__place_order__(self.livelineorder.stock, stock_order)
                order_placed.append(stock_order)
                self.logger.info('Stock order placed: %s' % stock_order)
            for options_order in options_pool.pool:
                self.__place_order__(self.livelineorder.options, options_order)
                order_placed.append(options_order)
                self.logger.info('Options order placed: %s' % options_order)

        finally:
            # save execution results
            now = self.market_time_now()
            day = now.format('YYYY-MM-DD')
            execution = {
                'timestamp': now.timestamp,
                'time': now.format('YYYY-MM-DD HH:mm:ss ZZ'),
                'order_placed': [o.to_dict() for o in order_placed]
            }
            self.execution_store.insert(
                day,
                strategy.name,
                execution)
            self.logger.debug('Execution saved: %s %s' % (day, strategy.name))
            self.logger.info('<< Finished strategy %s' % strategy.name)

    def __place_order__(self, executor: OrderExecutor, order: Order):
        try:
            executor.place_order(order.equity, order.quantity,
                                 order.price, order.order_type)
        except Exception as e:
            self.logger.error('Failed to place order <%s>' % order)
            raise e



import logging
from .order_executor import OrderExecutor
from liveline.model import Stock, Options


class LiveLineOrder:
    def __init__(self, stock: OrderExecutor[Stock], options: OrderExecutor[Options]):
        self._stock = stock
        self._options = options

        if self._stock is not None and self._stock.logger is None:
            self._stock.logger = logging.getLogger('LL/StockOrderExec')
        if self._options is not None and self._options.logger is None:
            self._options.logger = logging.getLogger('LL/OptionsOrderExec')

    @property
    def stock(self) -> OrderExecutor[Stock]:
        return self._stock

    @property
    def options(self) -> OrderExecutor[Options]:
        return self._options



import pymongo


class ExecutionDataStore:
    def __init__(self):
        pass

    def insert(self, day: str, strategy: str, execution: str):
        pass

    def get(self, day: str, strategy: str):
        pass


class MongoExecutionDataStore(ExecutionDataStore):
    """
    Save the executions of a strategy during a day

    Mongodb collection looks like:
    {
        'day': '2018-09-01',
        'strategy': 'buy FB',
        'executions': [
            {
                timestamp: 12345678,
                order_placed: [
                    {...}
                ]
            },
            {
                timestamp: 12345679,
                order_placed: []
            }
        ]
    }
    """

    def __init__(self, db: pymongo.database.Database):
        self.db: pymongo.database.Database = db

    def insert(self, day: str, strategy: str, execution: str):
        strategy_executions = self.db.executions.find_one({
            'day': day,
            'strategy': strategy
        })

        if strategy_executions is None:
            strategy_executions = {
                'day': day,
                'strategy': strategy,
                'executions': []
            }

        strategy_executions['executions'].append(execution)

        self.db.executions.update(
            {
                'day': day,
                'strategy': strategy
            }, 
            {
                '$set': {'executions': strategy_executions['executions']}
            }, 
            upsert=True)

    def get(self, day: str = None, strategy: str = None):
        filter = {}
        if day is not None: 
            filter['day'] = day
        if strategy is not None:
            filter['strategy'] = strategy

        return list(self.db.executions.find(filter, {'_id': False}))

    def clear(self, day: str = None, strategy: str = None):
        filter = {}
        if day is not None: 
            filter['day'] = day
        if strategy is not None:
            filter['strategy'] = strategy

        return self.db.executions.delete_many(filter)


import urllib


from flask import Blueprint, json, request


def build_overview_blueprint(ll):
    from liveline import LiveLine

    liveline: LiveLine = ll
    overview = Blueprint('overview', __name__)

    # pylint: disable=W0612
    @overview.route('/')
    def get_overview():
        base_url = request.host_url
        endpoints = {
            'strategies': urllib.parse.urljoin(base_url, '/strategy/'),
            'executions': urllib.parse.urljoin(base_url, '/execution/')
        }
        running = not liveline.paused
        executions = liveline.execution_store.get()
        last_execution_day = sorted([e['day']
                                     for e in executions], reverse=True)
        last_execution_day = last_execution_day[0] if len(last_execution_day) > 0 else ''
        last_executions = [e for e in executions if e['day'] == last_execution_day] 

        # filter out silent executions
        for execution in last_executions:
            execution['executions'] = [e for e in execution['executions'] if len(e['order_placed']) > 0]

        return json.jsonify({
            'running': running,
            'last_executions': last_executions,
            'endpoints': endpoints
        })

    return overview



from flask import Blueprint, json


def build_executions_blueprint(ll):
    from liveline import LiveLine

    liveline: LiveLine = ll
    executions = Blueprint('executions', __name__)

    # pylint: disable=W0612
    @executions.route('/')
    def get_all():
        executions = liveline.execution_store.get()
        return json.jsonify(executions)

    return executions


from .equity import Equity


class Position:
    def __init__(self):
        pass

    def to_dict(self):
        return {
            'equity': self.equity.to_dict(),
            'quantity': self.quantity,
            'average_cost': '%.2f' % self.average_cost
        }

    def __repr__(self):
        return "[%.0f x %s @ $%.2f]" % (self.quantity, self.equity, self.average_cost)

    @property
    def equity(self) -> Equity:
        pass

    @property
    def quantity(self) -> float:
        pass

    @property
    def average_cost(self) -> float:
        pass



from .portfolio import Portfolio


class Account:
    def __init__(self):
        pass

    @property
    def name(self) -> str:
        pass

    @property
    def buying_power(self) -> float:
        pass

    @property
    def total_positions_value(self) -> float:
        pass

    @property
    def day_trades_remaining(self) -> float:
        pass

    @property
    def max_funds_usable(self) -> float:
        pass

    @property
    def portfolio(self) -> Portfolio:
        pass



import logging
from .data_provider import DataProvider
from liveline.model import Stock, Options, Account


class LiveLineData:
    def __init__(self, account: Account, stock: DataProvider[Stock], options: DataProvider[Options]):
        self._account = account
        self._stock = stock
        self._options = options

        if self._stock is not None and self._stock.logger is None:
            self._stock.logger = logging.getLogger('LL/StockDataProvider')
        if self._options is not None and self._options.logger is None:
            self._options.logger = logging.getLogger('LL/OptionsDataProvider')

    @property
    def account(self) -> Account:
        return self._account

    @property
    def stock(self) -> DataProvider[Stock]:
        return self._stock

    @property
    def options(self) -> DataProvider[Options]:
        return self._options



import logging
from typing import Generic, TypeVar, List
from liveline.model import Order, Equity, OrderType, OrderStatus


E = TypeVar('E', bound=Equity)


# pylint: disable=E1136
class OrderExecutor(Generic[E]):
    def __init__(self): 
        self._logger = None

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger):
        self._logger = logger

    def place_order(self, equity: E, quantity: int, price: float, order_type=OrderType.MARKET) -> Order[E]:
        pass

    def cancel_order(self, order: Order[E]) -> Order[E]:
        pass


# pylint: disable=E1136
class PoolExecutor(OrderExecutor, Generic[E]):
    """
    This will not actually execut the orders, it will only store all orders
    in a order pool, which can be used to validate all orders.
    """

    def __init__(self):
        OrderExecutor.__init__(self)
        self._pool: List[Order] = list()

    @property
    def pool(self) -> List[Order]:
        return self._pool

    def place_order(self, equity: E, quantity: int, price: float, order_type=OrderType.MARKET) -> Order[E]:
        order = Order(equity, quantity, price, order_type=order_type)
        self._pool.append(order)
        return order

    def cancel_order(self, order: Order[E]) -> Order[E]:
        order.order_status = OrderStatus.CANCELED
        return order



import json
import time
from typing import List

import pymongo


class Context:
    def __init__(self, json_source=None, dict_source=None):
        if json_source is not None:
            self.__dict__ = json.loads(json_source)
        if dict_source is not None:
            self.__dict__ = dict_source

    def to_json(self) -> str:
        return json.dumps(self.__dict__)

    def __repr__(self):
        return self.to_json()

    def __getattr__(self, name):
        return None


class ContextDataStore:
    def __init__(self):
        pass

    def save(self, strategy_name: str, context: Context):
        pass

    def get_strategy_context(self, strategy_name: str) -> List[Context]:
        pass

    def get_strategy_latest_context(self, strategy_name: str) -> Context:
        pass

    def clear_strategy_context(self, strategy_name: str):
        pass


class MongoContextDataStore(ContextDataStore):
    def __init__(self, db: pymongo.database.Database):
        self.db: pymongo.database.Database = db

    def save(self, strategy_name: str, context: Context):
        data = {
            'context': context.to_json(),
            'strategy_name': strategy_name,
            'timestamp': int(time.time())
        }
        self.db.context.insert_one(data)

    def get_strategy_context(self, strategy_name: str) -> List[Context]:
        data = self.db.context.find({'strategy_name': strategy_name}) 
        return [Context(json_source=d['context']) for d in data]

    def get_strategy_latest_context(self, strategy_name: str) -> Context:
        results = self.db.context.find({'strategy_name': strategy_name}).sort(
            'timestamp', pymongo.DESCENDING).limit(1)
        results = list(results)

        if len(results) == 1:
            return Context(json_source=results[0]['context'])
        else:
            return None

    def clear_strategy_context(self, strategy_name: str):
        self.db.context.delete_many({'strategy_name': strategy_name})


from enum import Enum
from datetime import datetime
from typing import TypeVar, Generic


class Equity:
    def __init__(self):
        pass

    def to_dict(self):
        pass

    @property
    def name(self) -> str:
        pass

    @property
    def multipiler(self) -> float:
        pass

class Stock(Equity):
    def __init__(self, symbol):
        self._symbol = symbol

    def to_dict(self):
        return {'symbol': self.symbol}

    @property
    def name(self) -> str:
        return self.symbol

    def __repr__(self):
        return self.name

    @property
    def symbol(self) -> str:
        return self._symbol

    @property
    def multipiler(self):
        return 1.0


class ETF(Equity):
    def __init__(self):
        pass

    @property
    def name(self) -> str:
        pass

    @property
    def symbol(self) -> str:
        pass


OptionsUnderlyingEquity = TypeVar('OptionsUnderlyingEquity', Stock, ETF)


class OptionsType(Enum):
    CALL = 0
    PUT = 1

    def __str__(self):
        return self.name


# pylint: disable=E1136
class Options(Equity, Generic[OptionsUnderlyingEquity]):
    def __init__(self,
                 equity: OptionsUnderlyingEquity,
                 exp_date: datetime,
                 strike_price: float,
                 option_type: OptionsType):
        self._underlying_equity = equity
        self._exp_date = exp_date
        self._strike_price = strike_price
        self._option_type = option_type

    @property
    def name(self) -> str:
        return '%s %.2f %s %s' % (self.underlying_equity, 
                                  self.strike_price, 
                                  self.option_type, 
                                  self.expiration_date.strftime('%Y-%m-%d'))

    def to_dict(self):
        return {
            'underlying_equity': self.underlying_equity.to_dict(),
            'strike_price': '%.2f' % self.strike_price,
            'expiration_date': self.expiration_date.strftime('%Y-%m-%d'),
            'type': str(self.option_type)
        }

    def __repr__(self):
        return '<%s>' % self.name

    @property
    def multipiler(self):
        return 100.0

    @property
    def underlying_equity(self) -> OptionsUnderlyingEquity:
        return self._underlying_equity

    @property
    def expiration_date(self) -> datetime:
        return self._exp_date

    @property
    def strike_price(self) -> float:
        return self._strike_price

    @property
    def option_type(self) -> OptionsType:
        return self._option_type



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup
from setuptools import find_packages

setup(name='liveline',
      version='0.8',
      description='Python live trading framework',
      author='Daniel Wang',
      author_email='danielwpz@gmail.com',
      license='MIT',
      packages=find_packages(exclude=['tests*']),
      install_requires=[
          'arrow',
          'maya',
          'pandas',
          'pymongo',
          'pytz',
          'schedule',
          'flask',
          'pandas-market-calendars'
      ],
      zip_safe=False)



import logging
import pandas
from typing import Generic, TypeVar
from liveline.model import Equity, Quote


E = TypeVar('E', bound=Equity)


class DataProvider(Generic[E]):
    def __init__(self):
        self._logger = None

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @logger.setter
    def logger(self, logger: logging.Logger):
        self._logger = logger

    def get_quote(self, equity: E) -> Quote:
        pass

    def get_historicals(self, equity: E) -> pandas.DataFrame:
        pass 



from flask import Blueprint, json


def build_strategy_blueprint(ll):
    # we have to import LiveLine here because Engine is importing this module,
    # so it will be a circular import if we import LiveLine globally
    from liveline import LiveLine

    liveline: LiveLine = ll
    strategy_bp = Blueprint('strategy', __name__)

    # pylint: disable=W0612
    @strategy_bp.route('/')
    def list_strategies():
        strategies = liveline.get_strategies()
        return json.jsonify(list(strategies.keys()))

    return strategy_bp